#include "command_hotkeys.hpp"

#include <cstdarg>
#include "nlohmann/json.hpp"
#include "application.hpp"
#include "hotkey_system.hpp"
#include "hotkey_parser.hpp"
#include "json_utils.hpp"
#include "wc3/wc3props.hpp"
#include "command_overrides.hpp"
#include "wc3/game_ui.hpp"
#include "wc3/agile_data.hpp"

using namespace std;
using nlohmann::json;
using hotkey_system::Hotkey;
using hotkey_system::HotkeyCallback;
using hotkey_system::KeyModifier;

enum class SuperAction {
	PressButton,
	CastOnSelf,
	CastOnCursor,
	ToggleAutocast,
};

const char* SUPER_ACTION_NAMES[] = {
	"PressButton",
	"CastOnSelf",
	"CastOnCursor",
	"ToggleAutocast",
};

using ActionsTable = std::map<int, SuperAction>;

static const char* CONFIG_FILE_PATH = "./super_hotkey.json";

static vector<Hotkey> command_bar_hotkeys;
static vector<Hotkey> submenu_hotkeys;
static vector<Hotkey> inventory_bar_hotkeys;
static ActionsTable actions_table;
static Hotkey close_submenu_hotkey;
static Hotkey multiboard_toggle_hotkey;
static Hotkey super_mode_toggle_hotkey;
static Hotkey hide_button_hotkey;
static Hotkey reset_hidden_buttons_hotkey;
static bool super_mode_enabled = false;

static bool toggle_command_autocast(CCommandButtonData& command_data) {
	int command_index = get_command_index(
		command_data,
		MOUSE_BUTTON_FLAGS::RIGHT_MOUSE_BUTTON
	);
	if (command_index != -1) {
		return CGameUI::instance()->_select_mode->push_command(
			command_data.ability_id,
			command_data.order_id[command_index],
			command_data.order_flags[command_index],
			nullptr,
			nullptr,
			false
		);
	}

	return false;
}

static const uint32_t GHOST_IMAGE_TYPE_ID = '+wgh';

const uint32_t AGENT_CLASS_ID = '+agl';

static CAgent* get_agent_from_tag(const NetTag& tag) {
	CPresenceTagged* presence_tagged = tempest_thread::instance()->get_container(tag);
	if (presence_tagged && presence_tagged->ClassId == AGENT_CLASS_ID) {
		CAgentBaseAbs* container = static_cast<CAgentBaseAbs*>(presence_tagged);
		return container->value;
	}

	return nullptr;
}

static bool is_target_mode() {
	CGameUI* game_ui = CGameUI::instance();
	return (game_ui
		&& (game_ui->_current_input_mode == game_ui->_target_mode
			|| game_ui->_current_input_mode == game_ui->_build_mode
		)
	);
}

static bool activate_command(const CCommandButtonData& command_data) {
	int index = get_command_index(command_data, MOUSE_BUTTON_FLAGS::LEFT_MOUSE_BUTTON);
	CItem* item = (CItem*)get_agent_from_tag({
		command_data.item_presence[index],
		command_data.item_birth[index]
	});

	return CGameUI::instance()->_select_mode->push_command(
		command_data.ability_id,
		command_data.order_id[index],
		command_data.order_flags[index],
		item,
		nullptr,
		false
	);
}

static bool use_command_on_target(const CCommandButtonData& command_data, CSelectable* target, float x, float y) {
	CGameUI* game_ui = CGameUI::instance();

	if (is_target_mode()) {
		game_ui->close_current_input_mode();
	}

	if (!activate_command(command_data)) {
		return false;
	};

	if (!is_target_mode()) {
		return true;
	}

	// TODO: cursor should be within world frame.
	CGhostSpriteClickEvent event;
	event.point.x = x;
	event.point.y = y;
	event.mouse_button = MOUSE_BUTTON_FLAGS::LEFT_MOUSE_BUTTON;

	if (target) {
		if (AGILE_DATA::instance()->is_subtype(target->get_type_id(), GHOST_IMAGE_TYPE_ID)) {
			event.id = GHOST_SPRITE_CLICK_EVENT_ID;
			event.ghost_image = target;
			event.field34 = nullptr;
		}
		else {
			event.id = SPRITE_CLICK_EVENT_ID;
			event.widget = target;
			event.AllocatedHandleObjectLeaf_CSpriteUber_ = nullptr;
		}
	}
	else {
		event.id = TERRAIN_CLICK_EVENT_ID;
	}

	game_ui->_world_frame->emit(&event);

	if (is_target_mode()) {
		game_ui->close_current_input_mode();
	}

	return true;
}

static bool use_command_on_selected_unit(const CCommandButtonData& command_data) {
	CUnitSet* subgroup = CGameWar3::instance()->local_player()->selection->async_subgroup;
	if (subgroup && subgroup->items.count) {
		CUnit* unit = subgroup->items.first->value;
		return use_command_on_target(command_data, unit, unit->x, unit->y);
	}

	return false;
}

static void emulate_interface_click() {
	int context_handle = (int)get_wc3_prop(Wc3PropSlot::EVENT_CONTEXT_HANDLE);
	EvtContext* context = p_evt_context_table->get_item(context_handle);
	if (context) {
		Vector2f mouse_position = get_mouse_position_percents();

		EVENT_DATA_MOUSE event_data;
		event_data.mouse_key = LEFT_MOUSE_BUTTON;
		event_data.mouse_keys = LEFT_MOUSE_BUTTON;
		event_data.x = mouse_position.x;
		event_data.y = mouse_position.y;
		event_data.time = GetTickCount();
		emit_event(context, EVT_TYPE::MOUSE_DOWN, &event_data);

		event_data.mouse_keys = 0;
		event_data.time = GetTickCount();
		emit_event(context, EVT_TYPE::MOUSE_UP, &event_data);

		p_evt_context_table->unlock_item(context_handle);
	}
}

static bool use_command_on_cursor(const CCommandButtonData& command_data) {
	if (is_target_mode()) {
		CGameUI::instance()->close_current_input_mode();
	}

	if (activate_command(command_data)) {
		if (is_target_mode()) {
			emulate_interface_click();

			if (is_target_mode()) {
				CGameUI::instance()->close_current_input_mode();
				return false;
			}
		}

		return true;
	};

	return false;
}

static bool perform_action(SuperAction action, CCommandButton& button) {
	if (!button.command_data) {
		return false;
	}

	switch (action) {
		case SuperAction::PressButton:
			button.click(LEFT_MOUSE_BUTTON);
			return true;
		case SuperAction::CastOnSelf:
			return use_command_on_selected_unit(*button.command_data);
		case SuperAction::CastOnCursor:
			return use_command_on_cursor(*button.command_data);
		case SuperAction::ToggleAutocast:
			return toggle_command_autocast(*button.command_data);
     }

	 return false;
}

static bool parse_hotkeys_from_json_array(const json& array, const char* array_name, vector<Hotkey>& list) {
	int i = 0;

	for (json hotkey_field : array) {
		if (!json_check_type(hotkey_field, json::value_t::string, "%s[%d]", array_name, i)) {
			return false;
		}

		Hotkey hotkey;
		string hotkey_string = hotkey_field.get<string>();
		if (!try_parse_hotkey(hotkey_string.c_str(), hotkey)) {
			return false;
		}

		list.push_back(hotkey);
		i++;
	}

	return true;
}

static bool load_hotkey_field(const json& object, const char* field_name, Hotkey& hotkey) {
	string hotkey_string;
	return json_get_field(object, field_name, hotkey_string)
		&& try_parse_hotkey(hotkey_string.c_str(), hotkey)
		;
}

static bool load_hotkeys_field(const json& object, const char* field_name, vector<Hotkey>& list) {
	json hotkeys;
	return json_get_field(object, field_name, hotkeys)
		&& json_check_type(hotkeys, json::value_t::array, field_name)
		&& parse_hotkeys_from_json_array(hotkeys, field_name, list)
		;
}

static bool json_parse_hotkey_modifiers(const json& j_string, int& modifiers, const char* fmt, ...) {
	if (!try_parse_hotkey_modifiers(j_string.get<std::string>().c_str(), modifiers)) {
		char buffer[1024];
		va_list args;
		va_start(args, fmt);
		vsnprintf(buffer, sizeof(buffer), fmt, args);
		va_end(args);

		fprintf(
			stderr,
			"[%s] Failed to parse the hotkey modifiers from the config field «%s».\n",
			APPLICATION_NAME,
			buffer
		);

		return false;
	}

	return true;
}

static bool json_parse_super_action(const json& j_string, SuperAction& action, const char* fmt, ...) {
	const std::string& action_name = j_string.get<std::string>();

	for (int i = 0; i < std::size(SUPER_ACTION_NAMES); i++) {
		if (action_name == SUPER_ACTION_NAMES[i]) {
			action = (SuperAction)i;
			return true;
		}
	}

	char buffer[1024];
	va_list args;
	va_start(args, fmt);
	vsnprintf(buffer, sizeof(buffer), fmt, args);
	va_end(args);

	fprintf(
		stderr,
		"[%s] Failed to parse the config field «%s». Unknown action name: «%s».\n",
		APPLICATION_NAME,
		buffer,
		action_name.c_str()
	);

	return false;
}

static bool parse_actions_from_json_array(const json& array, const char* array_name, ActionsTable& actions) {
	int i = 0;

	for (json j_action : array) {
		if (!json_check_type(j_action, json::value_t::array, "%s[%d]", array_name, i)) {
			return false;
		}

		int modifiers;
		if (!json_check_index(j_action, 0, "%s[%d]", array_name, i)
			|| !json_check_type(j_action[0], json::value_t::string, "%s[%d]", array_name, i)
			|| !json_parse_hotkey_modifiers(j_action[0], modifiers, "%s[%d]", array_name, i)
		) {
			return false;
		}

		SuperAction action;
		if (!json_check_index(j_action, 1, "%s[%d]", array_name, i)
			|| !json_check_type(j_action[1], json::value_t::string, "%s[%d]", array_name, i)
			|| !json_parse_super_action(j_action[1], action, "%s[%d]", array_name, i)
		) {
			return false;
		}

		actions[modifiers] = action;

		i++;
	}

	return true;
}

static bool load_actions_field(const json& object, const char* field_name, ActionsTable& actions) {
	json j_actions;
	return json_get_field(object, field_name, j_actions, json::value_t::array)
		&& parse_actions_from_json_array(j_actions, field_name, actions)
		;
}

static bool load_config() {
	json root;

	return json_parse_file(CONFIG_FILE_PATH, root)
		&& json_check_type(root, json::value_t::object, "<root>")
		&& load_hotkeys_field(root, "command_bar_hotkeys", command_bar_hotkeys)
		&& load_hotkeys_field(root, "submenu_hotkeys", submenu_hotkeys)
		&& load_hotkeys_field(root, "inventory_bar_hotkeys", inventory_bar_hotkeys)
		&& load_actions_field(root, "actions", actions_table)
		&& load_hotkey_field(root, "close_submenu_hotkey", close_submenu_hotkey)
		&& load_hotkey_field(root, "multiboard_toggle_hotkey", multiboard_toggle_hotkey)
		&& load_hotkey_field(root, "super_mode_toggle_hotkey", super_mode_toggle_hotkey)
		&& load_hotkey_field(root, "hide_button_hotkey", hide_button_hotkey)
		&& load_hotkey_field(root, "reset_hidden_buttons_hotkey", reset_hidden_buttons_hotkey)
		;
}

template<typename Callable>
void for_each_command_button(Callable f) {
	CGameUI* game_ui = CGameUI::instance();
	if (!game_ui) {
		return;
	}

	for (int y = 0; y < game_ui->_command_bar->height; y++) {
		for (int x = 0; x < game_ui->_command_bar->width; x++) {
			CCommandButton* button = static_cast<CCommandButton*>(game_ui->_command_bar->rows[y][x]);
			if (button && !f(*button, x, y)) {
				return;
			};
		}
	}
}

static bool is_game_accepts_hotkeys(bool allow_target_mode = false) {
	CGameUI* game_ui = CGameUI::instance();
	return game_ui
		&& (
			(game_ui->_current_input_mode == game_ui->_select_mode)
			|| (game_ui->_current_input_mode == game_ui->_drag_select_mode)
			|| (game_ui->_current_input_mode == game_ui->_drag_scroll_mode)
			|| (allow_target_mode && (
					(game_ui->_current_input_mode == game_ui->_target_mode)
					|| (game_ui->_current_input_mode == game_ui->_build_mode)
				))
		)
		&& !game_ui->_chat_edit_bar->opened;
}

static CCommandButton* get_hovering_command_bar_button() {
	CCommandButton* result = nullptr;
	CSimpleTop* top = CSimpleTop::instance();

	if (top) {
		for_each_command_button([&] (CCommandButton& button, int x, int y) {
			if (&button == top->hovered_frame) {
				result = &button;
				return false;
			}
			else {
				return true;
			}
		});
	}

	return result;
}

static bool set_button_position(CCommandButton& button, int x, int y) {
	CGameUI* game_ui = CGameUI::instance();
	if (button.command_data
		&& x < game_ui->_command_bar->width
		&& y < game_ui->_command_bar->height
	) {
		set_command_position(*button.command_data, {x, y});
		game_ui->_command_bar->update();
		return true;
	}

	return false;
}

static bool set_button_position_by_index(CCommandButton& button, int index) {
	CGameUI* game_ui = CGameUI::instance();
	int x = index % game_ui->_command_bar->width;
	int y = index / game_ui->_command_bar->width;
	return set_button_position(button, x, y);
}

enum class PanelType: unsigned int {
	CommandBar,
	InventoryBar,
};

enum class SubmenuMode: unsigned int {
	Closed,
	Opened,
};

struct HotkeyParameter {
	unsigned int index: 30;
	PanelType panel_type: 1;
	SubmenuMode submenu_mode: 1;
};
static_assert(sizeof(HotkeyParameter) <= sizeof(void*));

CCommandButton* get_panel_button(PanelType panel_type, int button_index) {
	CGameUI* game_ui = CGameUI::instance();

	switch (panel_type) {
		case PanelType::CommandBar: {
			int x = button_index % game_ui->_command_bar->width;
			int y = button_index / game_ui->_command_bar->width;
			return game_ui->_command_bar->get_button(x, y);
		}
		case PanelType::InventoryBar: {
			CInventorySlots& slots = game_ui->_info_bar->inventory_bar->slots;
			return button_index < slots.size() ? slots[button_index].button : nullptr;
		}
	}

	return nullptr;
}

static SubmenuMode get_current_submenu_mode() {
	auto& submenus = CGameUI::instance()->_command_bar->submenu_items;
	return submenus.empty() ? SubmenuMode::Closed : SubmenuMode::Opened;
}

static bool on_command_bar_hotkey(Hotkey& hotkey, int extra_modifiers, void* parameter) {
	CGameUI* game_ui = CGameUI::instance();
	if (!game_ui) {
		return false;
	}

	if (is_game_accepts_hotkeys(true)) {
		HotkeyParameter* hotkey_parameter = reinterpret_cast<HotkeyParameter*>(&parameter);

		if ((hotkey_parameter->panel_type == PanelType::CommandBar)
			&& (get_current_submenu_mode() != hotkey_parameter->submenu_mode)
		) {
			return false;
		}

		if (super_mode_enabled) {
			CCommandButton* hovered_button = get_hovering_command_bar_button();
			return hovered_button && set_button_position_by_index(
				*hovered_button,
				hotkey_parameter->index
			);
		}
		else {
			auto action = actions_table.find(extra_modifiers & ~KeyModifier::SHIFT);
			if (action != actions_table.end()) {
				CCommandButton* button = get_panel_button(
					hotkey_parameter->panel_type,
					hotkey_parameter->index
				);
				if (button) {
					perform_action(action->second, *button);
					return true;
				}
			}
		}
	}

	return false;
}

static bool on_close_submenu_hotkey(Hotkey& hotkey, int extra_modifiers, void* parameter) {
	if (is_game_accepts_hotkeys()) {
		return CGameUI::instance()->_command_bar->close_submenu();
	}

	return false;
}

static bool on_toggle_multiboard(Hotkey& hotkey, int extra_modifiers, void* parameter) {
	CGameUI* game_ui = CGameUI::instance();
	if (game_ui && game_ui->multiboard) {
		game_ui->multiboard->set_minimized(!game_ui->multiboard->minimized);
		return true;
	}

	return false;
}

static bool on_super_mode_toggle_hotkey(Hotkey& hotkey, int extra_modifiers, void* parameter) {
	CGameUI* game_ui = CGameUI::instance();
	if (game_ui) {
		super_mode_enabled = !super_mode_enabled;

		static const char* messages[2] {
			"|cFFFF0000[%s] Super mode disabled.|r",
			"|cFF00FF00[%s] Super mode enabled.|r",
		};
		char buffer[256];
		snprintf(buffer, sizeof(buffer), messages[super_mode_enabled], APPLICATION_NAME);

		game_ui->_world_frame_chat_message->print(buffer);
		return true;
	}

	return false;
}

static bool on_hide_button_hotkey(Hotkey& hotkey, int extra_modifiers, void* parameter) {
	if (super_mode_enabled) {
		CCommandButton* button = get_hovering_command_bar_button();
		if (button && button->command_data) {
			set_command_hidden(*button->command_data, true);
			CGameUI::instance()->_command_bar->update();
			return true;
		}
	}

	return false;
}

static bool on_reset_hidden_buttons_hotkey(Hotkey& hotkey, int extra_modifiers, void* parameter) {
	CGameUI* game_ui = CGameUI::instance();
	if (game_ui) {
		reset_hidden_commands();
		game_ui->_command_bar->update();
		game_ui->_world_frame_chat_message->print("[Super hotkey] Hidden buttons list cleared.");
		return true;
	}

	return false;
}

static bool register_hotkeys(vector<Hotkey>& list, HotkeyCallback callback, PanelType panel_type, SubmenuMode submenu_mode = SubmenuMode::Closed) {
	for (unsigned int i = 0; i < list.size(); i++) {
		HotkeyParameter parameter {
			i,
			panel_type,
			submenu_mode,
		};

		bool success = hotkey_system::connect(
			list[i],
			callback,
			list[i].modifiers,
			*reinterpret_cast<void**>(&parameter)
		);
		if (!success) {
			return false;
		}
	}

	return true;
}

static bool unregister_hotkeys(vector<Hotkey>& list, HotkeyCallback callback, PanelType panel_type, SubmenuMode submenu_mode = SubmenuMode::Closed) {
	bool success = true;

	for (unsigned int i = 0; i < list.size(); i++) {
		HotkeyParameter parameter {
			i,
			panel_type,
			submenu_mode,
		};

		success = success && hotkey_system::disconnect(
			list[i],
			callback,
			*reinterpret_cast<void**>(&parameter)
		);
	}

	return success;
}

static bool register_all_hotkeys() {
	return register_hotkeys(command_bar_hotkeys, on_command_bar_hotkey, PanelType::CommandBar)
		&& register_hotkeys(submenu_hotkeys, on_command_bar_hotkey, PanelType::CommandBar, SubmenuMode::Opened)
		&& register_hotkeys(inventory_bar_hotkeys, on_command_bar_hotkey, PanelType::InventoryBar)
		&& hotkey_system::connect(close_submenu_hotkey, on_close_submenu_hotkey)
		&& hotkey_system::connect(multiboard_toggle_hotkey, on_toggle_multiboard)
		&& hotkey_system::connect(super_mode_toggle_hotkey, on_super_mode_toggle_hotkey)
		&& hotkey_system::connect(hide_button_hotkey, on_hide_button_hotkey)
		&& hotkey_system::connect(reset_hidden_buttons_hotkey, on_reset_hidden_buttons_hotkey)
		;
}

static bool unregister_all_hotkeys() {
	return unregister_hotkeys(command_bar_hotkeys, on_command_bar_hotkey, PanelType::CommandBar)
		&& unregister_hotkeys(submenu_hotkeys, on_command_bar_hotkey, PanelType::CommandBar, SubmenuMode::Opened)
		&& unregister_hotkeys(inventory_bar_hotkeys, on_command_bar_hotkey, PanelType::InventoryBar)
		&& hotkey_system::disconnect(close_submenu_hotkey, on_close_submenu_hotkey)
		&& hotkey_system::disconnect(multiboard_toggle_hotkey, on_toggle_multiboard)
		&& hotkey_system::disconnect(super_mode_toggle_hotkey, on_super_mode_toggle_hotkey)
		&& hotkey_system::disconnect(hide_button_hotkey, on_hide_button_hotkey)
		&& hotkey_system::disconnect(reset_hidden_buttons_hotkey, on_reset_hidden_buttons_hotkey)
		;
}

bool command_hotkeys_initialize() {
	set_default_hotkeys_disabled(true);
	return load_config() && register_all_hotkeys();
}

bool command_hotkeys_finalize() {
	set_default_hotkeys_disabled(false);
	return unregister_all_hotkeys();
}
