#include "hotkey_parser.hpp"

#include <cstdio>
#include <fstream>
#include <map>
#include "application.hpp"
#include <boost/algorithm/string.hpp>
#include "winuser.h"
#include "hotkey_system.hpp"
#include "nlohmann/json.hpp"
#include "json_utils.hpp"

using namespace std;
using json = nlohmann::json;
using namespace boost::algorithm;

static const char* KEY_CODES_FILE_PATH = "./super_hotkey.key_codes.json";

static map<string, DWORD> key_to_code_table;
static map<DWORD, hotkey_system::KeyModifier> key_code_to_modifier_table {
	{ VK_MENU, hotkey_system::KeyModifier::ALT },
	{ VK_CONTROL, hotkey_system::KeyModifier::CTRL },
	{ VK_SHIFT, hotkey_system::KeyModifier::SHIFT },
	{ VK_LWIN, hotkey_system::KeyModifier::WIN },
	{ VK_RWIN, hotkey_system::KeyModifier::WIN },
};

bool key_name_to_key_code(const char* key_name, DWORD& key_code) {
	string key_name_lower_case(key_name);
	to_lower(key_name_lower_case);

	auto it = key_to_code_table.find(key_name_lower_case);
	if (it == key_to_code_table.end()) {
		fprintf(
			stderr,
			"[%s] Key code with name «%s» not found.\n",
			APPLICATION_NAME,
			key_name
		);

		return false;
	}

	key_code = it->second;
	return true;
}

bool key_code_to_modifier(DWORD key_code, hotkey_system::KeyModifier& modifier) {
	auto it = key_code_to_modifier_table.find(key_code);
	if (it == key_code_to_modifier_table.end()) {
		fprintf(
			stderr,
			"[%s] Key modifier with virtual key code 0x%.2X not found.\n",
			APPLICATION_NAME,
			key_code
		);

		return false;
	}

	modifier = it->second;
	return true;
}

bool key_names_to_key_codes(const vector<string>& key_names, vector<DWORD>& key_codes) {
	for (auto key_name : key_names) {
		DWORD key_code;
		if (!key_name_to_key_code(key_name.c_str(), key_code)) {
			return false;
		}

		key_codes.push_back(key_code);
	}

	return true;
}

bool try_parse_hotkey_modifiers(const char* modifiers_string, int& modifiers) {
	vector<string> key_names;
	boost::split(key_names, modifiers_string, boost::is_any_of(" "));
	if ((key_names.size() == 1) && (key_names[0] == "")) {
		modifiers = 0;
		return true;
	}

	vector<DWORD> key_codes;
	return key_names_to_key_codes(key_names, key_codes)
		&& key_codes_to_modifiers(key_codes, modifiers);
}

bool try_parse_hotkey(const char* hotkey_string, hotkey_system::Hotkey& hotkey) {
	vector<string> key_names;
	split(key_names, hotkey_string, is_any_of(" "));
	if (!key_names.size()) {
		fprintf(stderr, "Failed to parse hotkey from empty string.");
		return false;
	}

	vector<DWORD> key_codes;
	if (!key_names_to_key_codes(key_names, key_codes)) {
		return false;
	}

	hotkey.key = key_codes[key_codes.size() - 1];
	hotkey.modifiers = 0;

	if (key_codes.size() > 1) {
		auto modifier_codes = boost::make_iterator_range(key_codes.begin(), key_codes.end() - 1);
		if (!key_codes_to_modifiers(modifier_codes, hotkey.modifiers)) {
			return false;
		}
	}

	return true;
}

static bool load_key_codes() {
	ifstream f(KEY_CODES_FILE_PATH);
	if (!f) {
		fprintf(stderr, "[%s] Failed to open the key codes file at «%s».\n", APPLICATION_NAME, KEY_CODES_FILE_PATH);
		return false;
	}

	json root;
	if (!json_parse_file(KEY_CODES_FILE_PATH, root)
		|| !json_check_type(root, json::value_t::object, "<root>")
	) {
		return false;
	}

	for (auto& [name, value] : root.items()) {
		if (!json_check_type(value, json::value_t::number_unsigned, name.c_str())) {
			return false;
		}

		string lower_name(name);
		to_lower(lower_name);

		auto code = value.get<DWORD>();

		if (key_to_code_table.find(lower_name) != key_to_code_table.end()) {
			fprintf(stderr, "[%s] Warning: duplicate entry with key «%s».\n", APPLICATION_NAME, lower_name.c_str());
		}

		key_to_code_table[lower_name] = code;
	}

	return true;
}

bool hotkey_parser_initialize() {
	return load_key_codes();
}
