#include "hotkey_system.hpp"

#include <cstdio>
#include <unordered_map>

#include <winuser.h>
#include <errhandlingapi.h>

#include "application.hpp"

namespace hotkey_system {

	size_t Hotkey::operator == (const Hotkey& other) const {
		return (key == other.key) && (modifiers == other.modifiers);
	}

	size_t Hotkey::operator != (const Hotkey& other) const {
		return !(*this == other);
	}

	struct HotkeyHandler {
		Hotkey hotkey;
		HotkeyCallback callback;
		int modifiers_mask;
		void* parameter;

		bool accepts(const Hotkey& hotkey) const {
			return (
				(this->hotkey.key == hotkey.key)
				&& ((hotkey.modifiers & modifiers_mask) == this->hotkey.modifiers)
			);
		}
	};

	static HHOOK keyboard_hook = NULL;
	static std::unordered_map<DWORD, std::vector<HotkeyHandler>> handlers_registry;

	static bool is_key_pressed(DWORD key) {
		return GetKeyState(key) & (1 << 15);
	}

	static int get_pressed_modifiers() {
		int result = 0;

		if (is_key_pressed(VK_MENU)) { result |= KeyModifier::ALT; };
		if (is_key_pressed(VK_CONTROL)) { result |= KeyModifier::CTRL; };
		if (is_key_pressed(VK_SHIFT)) { result |= KeyModifier::SHIFT; };
		if (is_key_pressed(VK_LWIN)) { result |= KeyModifier::WIN; };
		if (is_key_pressed(VK_RWIN)) { result |= KeyModifier::WIN; };

		return result;
	}

	bool emit_key_event(DWORD key, int modifiers) {
		Hotkey hotkey { key, modifiers };

		auto it = handlers_registry.find(hotkey.key);
		if (it != handlers_registry.end()) {
			for (const HotkeyHandler& handler : it->second) {
				int extra_modifiers = hotkey.modifiers & ~handler.modifiers_mask;
				if (handler.accepts(hotkey)
					&& handler.callback(hotkey, extra_modifiers, handler.parameter)
				) {
					return true;
				};
			}
		}

		return false;
	}

	bool connect(Hotkey& hotkey, HotkeyCallback callback, int modifiers_mask, void* parameter) {
		handlers_registry[hotkey.key].push_back({hotkey, callback, modifiers_mask, parameter});
		return true;
	}

	bool disconnect(Hotkey& hotkey, HotkeyCallback callback, std::optional<void*> parameter) {
		auto it = handlers_registry.find(hotkey.key);
		if (it == handlers_registry.end()) {
			return false;
		}

		std::vector<HotkeyHandler>& handlers = it->second;
		for (int i = 0; i < handlers.size(); i++) {
			if (handlers[i].callback == callback
				&& handlers[i].hotkey.modifiers == hotkey.modifiers
				&& (!parameter.has_value() || handlers[i].parameter == parameter.value())
			) {
				handlers.erase(handlers.begin() + i);
				return true;
			}
		}

		return false;
	}

	static LRESULT CALLBACK KeyboardProc(int code, WPARAM wParam, LPARAM lParam) {
		if (code == HC_ACTION) {
			bool is_key_pressed = (lParam >> 31) == 0;
			if (is_key_pressed) {
				bool handled = emit_key_event(wParam, get_pressed_modifiers());
				if (handled) {
					return 1;
				}
			};
		}

		return CallNextHookEx(keyboard_hook, code, wParam, lParam);
	}

	bool attach_to_window(HWND window) {
		detach_from_current_window();
		
		keyboard_hook = SetWindowsHookExA(
			WH_KEYBOARD,
			KeyboardProc,
			NULL,
			GetWindowThreadProcessId(window, nullptr)
		);
		if (keyboard_hook == NULL) {
			fprintf(
				stderr,
				"[%s] Failed to create keyboard hook. Error = %d\n",
				APPLICATION_NAME,
				GetLastError()
			);
			return false;
		}

		return true;
	}

	bool detach_from_current_window() {
		if (keyboard_hook != NULL) {
			bool success = UnhookWindowsHookEx(keyboard_hook);
			keyboard_hook = NULL;
			return success;
		}

		return true;
	}

	bool initialize() {
		return true;
	}

	bool finalize() {
		return detach_from_current_window();
	}
}
