#include "json_utils.hpp"

#include "application.hpp"
#include <cstdarg>
#include <cstdio>
#include <fstream>

bool json_parse_file(const char* file_path, json& root) {
	std::ifstream f(file_path);
	if (!f) {
		fprintf(stderr, "[%s] Failed to open config file at «%s».\n", APPLICATION_NAME, file_path);
		return false;
	}

	try {
		root = json::parse(f);
	}
	catch (json::parse_error& e) {
		fprintf(
			stderr,
			"[%s] Failed to parse «%s»: %s\n",
			APPLICATION_NAME,
			file_path,
			e.what()
		);

		return false;
	}

	return true;
}

const char* json_type_name(json::value_t type)
{
	switch (type) {
		case json::value_t::null:
			return "null";
		case json::value_t::object:
			return "object";
		case json::value_t::array:
			return "array";
		case json::value_t::string:
			return "string";
		case json::value_t::boolean:
			return "boolean";
		case json::value_t::binary:
			return "binary";
		case json::value_t::discarded:
			return "discarded";
		case json::value_t::number_integer:
			return "number";
		case json::value_t::number_unsigned:
			return "unsigned number";
		case json::value_t::number_float:
			return "float";
		default:
			return "unknown";
	}
}

bool json_check_type(const json& data, json::value_t type, const char* fmt, ...) {
	if (data.type() != type) {
		char name_buffer[1024];

		std::va_list args;
		va_start(args, fmt);
		vsnprintf(name_buffer, sizeof(name_buffer), fmt, args);
		va_end(args);

		fprintf(
			stderr,
			"[%s] Config field «%s» has wrong type: expected «%s» but got «%s».\n",
			APPLICATION_NAME,
			name_buffer,
			json_type_name(type),
			json_type_name(data.type())
		);

		return false;
	}

	return true;
}

bool json_check_index(const json& array, int index, const char* fmt, ...) {
	if (index >= array.size()) {
		char buffer[1024];
		std::va_list args;
		va_start(args, fmt);
		vsnprintf(buffer, sizeof(buffer), fmt, args);
		va_end(args);

		fprintf(
			stderr,
			"[%s] Index %d is out of bounds. Length of «%s» = %d.\n",
			APPLICATION_NAME,
			index,
			buffer,
			array.size()
		);
		return false;
	}

	return true;
}

bool json_get_field(const char* array_name, const json& array, int index, json& field) {
	if (json_check_index(array, index, array_name)) {
		field = array[index];
		return true;
	}

	return false;
}

bool json_get_field(const char* array_name, const json& array, int index, json& field, json::value_t type) {
	return json_get_field(array_name, array, index, field)
		&& json_check_type(field, type, "%s[%d]", array_name, index)
		;
}

bool json_get_field(const json& object, const char* field_name, json& field) {
	auto it = object.find(field_name);
	if (it == object.end()) {
		fprintf(stderr, "[%s] Field «%s» not found.\n", APPLICATION_NAME, field_name);
		return false;
	}

	field = *it;
	return true;
}

bool json_get_field(const json& object, const char* field_name, json& field, json::value_t type) {
	return json_get_field(object, field_name, field)
		&& json_check_type(field, type, field_name)
		;
}
